/* 1949, Wed 16 Feb 00

   CYGWIN_NET.C: Cygwin support routines for libpcap
      These provide the interface to WinDump's packet.dll
      Having them here (rather than in pcap-cygwd.c) avoids
      networking conflicts between Cygwin and Windows.

      Also, cygwin didn't have snprintf(); we provide one here.

   Copyright (C) 1992-2002 by Nevil Brownlee,
   CAIDA & The University of Auckland */


#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <sys/types.h>

#include <windows.h>
#include "packet32.h"
#include "ntddndis.h"

#define CYGWD_EXT
#include "pcap-cygwd.h"

/* libpcap has a symbolic link net -> bpf/net, and bpf.h is there,
      with the defines for DLT_
   But we can't just include it, because it conflicts with 
      the bpf_xxx defines in packet32.h
   Instead we just copy the defines we need ... */

#define DLT_NULL	0	/* no link-layer encapsulation */
#define DLT_EN10MB	1	/* Ethernet (10Mb) */
#define DLT_EN3MB	2	/* Experimental Ethernet (3Mb) */
#define DLT_AX25	3	/* Amateur Radio AX.25 */
#define DLT_PRONET	4	/* Proteon ProNET Token Ring */
#define DLT_CHAOS	5	/* Chaos */
#define DLT_IEEE802	6	/* IEEE 802 Networks */
#define DLT_ARCNET	7	/* ARCNET */
#define DLT_SLIP	8	/* Serial Line IP */
#define DLT_PPP		9	/* Point-to-point Protocol */
#define DLT_FDDI	10	/* FDDI */
#define DLT_ATM_RFC1483	11	/* LLC/SNAP encapsulated atm */
#define DLT_RAW		12	/* raw IP */
#define DLT_SLIP_BSDOS	13	/* BSD/OS Serial Line IP */
#define DLT_PPP_BSDOS	14	/* BSD/OS Point-to-point Protocol */

				/* Extra types added by WinDump */
#define DLT_EN100MB	100	/* Ethernet (100Mb) */
#define DLT_PPP_WIN32	101	/* Win32 dial up connection */


/* Structure to keep info about WinDump adapters.
   This is indexed by fd from the pcap_t structure */

struct cygwd_info {
   char *if_name;
   ADAPTER *adapter;
   LPPACKET Packet;
   };
struct cygwd_info cygwd_if[MX_INTERFACES];
int n_open_adapters = 0;

int find_interfaces(void)
{
   DWORD dwVersion, dwWindowsMajorVersion;
#define NB_SZ  500
   char nameasc[NB_SZ];  /* ASCII for Win9x */
   WCHAR nameunc[NB_SZ];  /* Unicode for NT */
   ULONG nbsz = NB_SZ;
   int j, k, len, u;

   dwVersion=GetVersion();
   dwWindowsMajorVersion =  (DWORD)(LOBYTE(LOWORD(dwVersion)));
   if (dwVersion >= 0x80000000 && dwWindowsMajorVersion >= 4) {
      PacketGetAdapterNames(nameasc, &nbsz);
      for (cyg_interfaces = j = k = 0; ; ++k) {  /* Win9x */
         if (nameasc[k] == ' ' || nameasc[k] == '\0') {
            if (k == j)  /* No more interfaces */
               return cyg_interfaces;
            nameasc[k] = '\0';
            len = k+1 - j;  /* Including trailing null */
            if_names[cyg_interfaces] = (char *)malloc(len);
            strcpy(if_names[cyg_interfaces++], &nameasc[j]);
            j = k+1;  /* Ready for next name */
            }
         }
      }
   else {  /* Windows NT */
      PacketGetAdapterNames((PSTR)nameunc, &nbsz); 
      for (cyg_interfaces = j = k = 0; ; ++k) {
         if (nameunc[k] == '\0') {  /* Unicode null */
            if (k == j)  /* No more interfaces */
               return cyg_interfaces;
            nameunc[k] = '\0';
            len = k+1 - j;  /* Including trailing null */
            if_names[cyg_interfaces] = (char *)malloc(len);
            for (u = j; u != len-1; ++u, ++j)  /* Low bytes of unicode => ASCII */
               if_names[cyg_interfaces][u] = nameunc[j];
            if_names[cyg_interfaces][len-1] = '\0';
            j = k+1;  /* Ready for next name */
            }
         }
      }
   }


int cygwd_setbuff(int fd, int dim)  /* Set packet.dll buffer size */
{
   return PacketSetBuff(cygwd_if[fd].adapter, dim);
   }

int cygwd_stats(int fd, struct bpf_stat *ps)
{
   return PacketGetStats(cygwd_if[fd].adapter, ps);
   }

int cygwd_have_adaptor(int fd)
{
   return cygwd_if[fd].adapter != NULL;
   }

int cygwd_setfilter(int fd, struct bpf_program *fp)
{
   return PacketSetBpf(cygwd_if[fd].adapter, fp);
   }

int cygwd_open_adapter(char *device)
{
   int fd;
   if ((cygwd_if[n_open_adapters].adapter = PacketOpenAdapter(device)) == NULL)
      return -1;
   else {
      fd = n_open_adapters++;      
      cygwd_if[fd].if_name = device;
      cygwd_if[fd].Packet = NULL;
      return fd;
      }
   }

void cygwd_set_hwfilter(int fd, int promisc)
{
   PacketSetHwFilter(cygwd_if[fd].adapter,
      promisc ?  NDIS_PACKET_TYPE_PROMISCUOUS : NDIS_PACKET_TYPE_DIRECTED);
   }

int cygwd_get_nettype(int fd, u_int *linktype)
{
   NetType type;
   if (PacketGetNetType(cygwd_if[fd].adapter,&type) == FALSE)
      return 0;
   else {
      /* Set the linktype.  Only Ethernet and PPP are recognized */
      switch (type.LinkType) {
      case NdisMediumWan:  /* PPP connection in win9x */
         *linktype = DLT_PPP_WIN32;  
         break;
      case NdisMedium802_3:
         if (type.LinkSpeed == 100000000)
            *linktype = DLT_EN100MB;
         else if (type.LinkSpeed == 10000000)
            *linktype = DLT_EN10MB;
         else
            *linktype = DLT_PPP_WIN32;  /* PPP connection in winNTx (!) */
         break;
      default:
         *linktype = DLT_EN10MB;  /* Assume it's Ethernet */
         }
      return 1;
      }
   }

int cygwd_init_adaptor(int fd, char *buf, int sz_buf, int to_ms)
{
   /* Allocate packet structure used during the capture */
   if ((cygwd_if[fd].Packet = PacketAllocatePacket()) == NULL)
      return CYF_PAP;

   PacketInitPacket(cygwd_if[fd].Packet, buf,sz_buf);

   /* Allocate the buffer in the driver */
   if (!PacketSetBuff(cygwd_if[fd].adapter, WDifBufSize))
      return CYF_PSB;

   if (!PacketSetReadTimeout(cygwd_if[fd].adapter, to_ms))
      return CYF_PST;

   return CYF_POK;
   }

void cygwd_close_adaptor(int fd)
{
   if (cygwd_if[fd].adapter) {
      PacketFreePacket(cygwd_if[fd].Packet);
      PacketSetHwFilter(cygwd_if[fd].adapter,0);
      PacketCloseAdapter(cygwd_if[fd].adapter);
      }
   }

int cygwd_receive_packet(int fd, int *cc)
{
   /* TRUE is for blocking IO
      FALSE -> must use PacketWaitPacket (which blocks) */
   if (!PacketReceivePacket(cygwd_if[fd].adapter, cygwd_if[fd].Packet, TRUE))
      return 0;
   else {
      *cc = cygwd_if[fd].Packet->ulBytesReceived;
      return 1;
      }
   }

int snprintf(char *str, size_t size, const char *format, ...)
{
   char msg[300];  /* Hope this is big enough! */
   int r;  va_list ap;

   va_start(ap, format);
   vsprintf(msg, format, ap);
   va_end(ap);
   if ((r = strlen(msg)) > sizeof(msg)) return 0;
   else {
      strcpy(str, msg);  return r;
      }
   }

int vsnprintf(char *str, size_t size, const char *format, va_list ap)
{
   char msg[300];  /* Hope this is big enough! */
   int r;

   vsprintf(msg, format, ap);
   va_end(ap);
   if ((r = strlen(msg)) > sizeof(msg)) return 0;
   else {
      strcpy(str, msg);  return r;
      }
   }
